#!/usr/bin/env python

# Copyright (C) 2017 Lenna X. Peterson, Daisuke Kihara, and Purdue University
# This file is part of Path-LZerD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import argparse
import errno
import glob
import inspect
import logging
import os
import shutil
import subprocess

import pandas as pd

import shared

script_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))


# Key: score
# Value: required item in PATHS.ini
possible_scores = dict(
    dfire="goap_path",
    goap="goap_path",
    itscore="itscorepro_cmd",
    opuspsp="opuspsp_path",
    soappp=["modeller_cmd", "soappp_cmd"],
    shapescore=None,
    molmech="lzerd_path",
    sum=None,
)

# Specify score filename and sort order if not matching default
score_files = dict(
    shapescore=dict(fmt="{0}-{1}.out", ascending=False),
    dfire=dict(fmt="{0}-{1}_goap.txt", ascending=True),
)

def get_score_dict(scorename):
    score_dict = score_files.get(scorename)
    if score_dict is None:
        score_dict = dict(fmt="{0}-{1}_%s.txt" % scorename,
                          ascending=True)
    return score_dict


class ScoreSharedError(RuntimeError):
    """Exception for class ScoreShared"""
    def __init__(self, message):
        super(ScoreSharedError, self).__init__(message)


class ScoreInterface(object):

    default_wd = os.getcwd()
    default_pattern = "*.pdb"
    default_output_file = "scores.txt"

    @classmethod
    def get_pdbs(cls, directory=None, input=None, pattern=None, wd=None):

        if pattern is None:
            cls.default_pattern
        if wd is None:
            wd = cls.default_wd

        filenames = list()
        if input is not None:
            if isinstance(input, str):
                if "*" in input:
                    filenames.extend(glob.glob(input))
                else:
                    filenames.append(input)
            else:
                filenames.extend(input)
        if directory is not None:
            filenames.extend(glob.glob(os.path.join(directory, pattern)))
        if not filenames:
            filenames = glob.glob(os.path.join(wd, "decoys", pattern))
            if not filenames:
                raise ScoreSharedError("No files were found")

        # check for bad hydrogen atoms
        with open(filenames[0]) as ih:
            for line in ih:
                # Reject atom lines where 16 is not blank (altloc)
                # and 12 is blank (first character of 4-letter atom name)
                # and 13 is not "C" or "N" (e.g. eta carbon)
                # and 14 is "H" (misaligned hydrogen)
                if line[:6] == "ATOM  " and \
                   line[16] != " " and \
                   line[12] == " " and \
                   line[13] not in ("C", "N") and \
                   line[14] == "H":
                    raise ScoreSharedError("Hydrogen atom '%s' in %s is misaligned. Remove hydrogen atoms or re-generate decoys." % (line[12:17], filenames[0]))

        return filenames

    @classmethod
    def get_relative_pdbs(cls, files):
        file_path = None
        rel_files = list()
        for f in files:
            filedirname, filebasename = os.path.split(f)
            if filedirname == "":
                filedirname = os.path.abspath(filedirname)
            rel_files.append(filebasename)
            if file_path is None:
                file_path = filedirname
            else:
                if filedirname != file_path:
                    raise ScoreSharedError("Cannot run files from multiple directories: %s %s", filedirname, file_path)

        return file_path, rel_files

    @classmethod
    def _default_parser(cls):
        parser = argparse.ArgumentParser(add_help=False,
                                         description="Score files")
        parser.add_argument("input", nargs="*",
                           help="Input file(s)")
        parser.add_argument("-d", "--directory",
                           help="Directory of input files")
        parser.add_argument("-p", "--pattern", default=cls.default_pattern,
                           help="File pattern in directory (default '{0}')".format(cls.default_pattern))
        parser.add_argument("-w", "--wd", default=cls.default_wd,
                           help="Working directory (where score file is written)")
        parser.add_argument("-o", "--output_file", default=cls.default_output_file,
                           help="Name of score file (default '{0}')".format(cls.default_output_file))
        parser.add_argument("-v", "--verbose", action="store_true",
                       help="Show debugging statements")
        return parser
    
    @classmethod
    def _get_parser(cls):
        parent_parser = cls._default_parser()
        parser = argparse.ArgumentParser(parents=[parent_parser])
        return parser

    @classmethod
    def commandline(cls, module_args=None):
        a = cls._get_parser()

        args = a.parse_args(module_args)
        kwargs = vars(args)

        level = logging.INFO
        if kwargs.pop("verbose"):
            level = logging.DEBUG
        logging.basicConfig(level=level)
        logging.debug(args)

        c = cls(**kwargs)
        return c


class GoapScoreError(RuntimeError):
    """Exception for class GoapScore"""
    def __init__(self, message):
        super(GoapScoreError, self).__init__(message)


class GoapScore(ScoreInterface):

    default_output_file = "goap_score.txt"
    default_max_score = -100

    def __init__(self, goap_path, input=None, directory=None, pattern=None, wd=None, output_file=None, max_score=None):
        "Constructor"

        self.goap_location = goap_path
        goap_bin = os.path.join(self.goap_location, "goap")
        if shared.missing(goap_bin):
            raise GoapScoreError("{0} not found".format(goap_bin))
        self.goap_bin = goap_bin

        if wd is None:
            wd = self.default_wd

        if pattern is None:
            pattern = self.default_pattern

        filenames = self.get_pdbs(input=input, directory=directory,
                                  pattern=pattern, wd=wd)

        if output_file is None:
            output_file = os.path.join(wd, self.default_output_file)

        output_dir = os.path.abspath(os.path.dirname(output_file))
        if not os.path.isdir(output_dir):
            raise GoapScoreError("Directory %s does not exist" % output_dir)

        print wd

        if max_score is None:
            max_score = self.default_max_score

        # TODO check whether it's done first
        # Check score magnitude of first non-failing file
        #for f in filenames:
            #test_result = self.run_goap(files=[f], output_file="/dev/null")
            #if test_result:
                #parts = test_result[0].split()
                #goap = float(parts[2])
                #if goap > max_score:
                    #raise GoapScoreError("{0} has score over {1}, add hydrogens".format(filenames[0], max_score))
                #break
            #elif len(filenames) == 1:
                #raise GoapScoreError("1 file, which failed")
        self.run_goap(files=filenames, output_file=output_file)
        logging.info("Goap job complete.")
        self.output_file = output_file

    def run_goap(self, files, output_file):

        output_dir = os.path.abspath(os.path.dirname(output_file))
        if not os.path.isdir(output_dir):
            raise GoapScoreError("Directory %s does not exist" % output_dir)

        if not files:
            logging.warning("No files")
            return

        if os.path.isfile(output_file) and os.path.getsize(output_file):
            goap_file_set = set()
            with open(output_file) as ih:
                for line in ih:
                    index, filename, a, b, c = line.split()
                    goap_file_set.add(filename)
            if goap_file_set == set(os.path.basename(f) for f in files):
                logging.info("Job complete")
                return

        file_path, rel_files = self.get_relative_pdbs(files)
        #file_path = None

        #rel_files = list()
        #for f in files:
            #filedirname, filebasename = os.path.split(f)
            #if filedirname == "":
                #filedirname = os.path.abspath(filedirname)
            #rel_files.append(filebasename)
            #if file_path is None:
                #file_path = filedirname
            #else:
                #if filedirname != file_path:
                    #raise GoapScoreError("Cannot run files from multiple directories: %s %s", filedirname, file_path)

        if file_path is None:
            raise GoapScoreError("Need absolute path of files")

        if not os.path.isfile(output_file) or not os.path.getsize(output_file):

            result = list()
            error_files = list()
            todo_files = list(rel_files)
            #logging.debug(todo_files)
            while todo_files:
                lines = self._goap(file_path=file_path, rel_files=todo_files)
                #logging.debug(len(lines))
                if len(lines) + len(error_files) + len(result) == len(rel_files):
                    todo_files = list()
                else:
                    error_index = len(lines)
                    error_files.append(todo_files[error_index])
                    todo_files = todo_files[error_index + 1:]
                    #logging.debug(todo_files)
                result.extend(lines)

            with open(output_file, "w") as oh:
                for line in result:
                    oh.write(line + "\n")
            return result

    @classmethod
    def load(cls, filename):
        names = ("index", "model.pdb", "goap", "dfire", "goap_ag")
        types = (str, str, float, float, float)
        kwargs = dict(names=names,
                      dtype=zip(names, types),
                      delim_whitespace=True)
        mydf = pd.read_csv(filename, **kwargs)
        del mydf["index"]
        del mydf["goap_ag"]
        mydf['model'] = mydf.pop("model.pdb").str.replace(".pdb", "")
        return mydf

    def _goap(self, file_path, rel_files):
        #header = [self.goap_location, file_path]
        header = [self.goap_location]
        inp_str = "\n".join(header + rel_files)

        if logging.getLogger().isEnabledFor(logging.DEBUG):
            with open("testgoap.inp", "w") as oh:
                oh.write(inp_str)

        with shared.CHDIR(file_path):
            proc = subprocess.Popen([self.goap_bin],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            out, err = proc.communicate(inp_str)
        lines = out.splitlines()
        if proc.returncode:
            logging.error("Goap returned %s", proc.returncode)
            logging.debug(err)
        return lines

    @classmethod
    def _get_parser(cls):
        parent_parser = cls._default_parser()
        parser = argparse.ArgumentParser(parents=[parent_parser])
        parser.add_argument("-m", "--max_score", default=cls.default_max_score, type=float,
                            help="Maximum allowed score (scores around -30 typically mean missing hydrogens)")
        return parser


class ITScoreError(RuntimeError):
    """Exception for class ITScore"""
    def __init__(self, message):
        super(ITScoreError, self).__init__(message)


class ITScore(ScoreInterface):

    default_output_file = "scores.itscore"

    def __init__(self, itscorepro_cmd, input=None, directory=None, pattern=None, wd=None, output_file=None):
        """CONSTRUCTOR"""

        if shared.missing(itscorepro_cmd):
            raise ITScoreError("{0} missing".format(itscorepro_cmd))
        self.itscore_bin = itscorepro_cmd

        if wd is None:
            wd = self.default_wd

        if pattern is None:
            pattern = self.default_pattern

        filenames = self.get_pdbs(input=input, directory=directory,
                                  pattern=pattern, wd=wd)

        if output_file is None:
            output_file = self.default_output_file
        output_file = os.path.join(wd, output_file)
        logging.info(wd)
        logging.info("%s files to ITScore", len(filenames))
        fn = self.run(pdbfiles=filenames, fn=output_file)
        self.output_file = fn

    def run(self, pdbfiles, fn=None):
        if fn is None:
            fn = self.default_output_file
        errors = list()
        # List models already in file
        filemodel_dict = dict()
        if os.path.isfile(fn) and os.path.getsize(fn):
            with open(fn) as ih:
                for line in ih:
                    model, score = line.split()
                    filemodel_dict[model] = score
        # List models in PDB argument
        pdb_mdl_dict = {os.path.splitext(os.path.basename(m))[0]: m for m in pdbfiles}
        # Find PDBs missing from file
        to_score = set(pdb_mdl_dict.keys()).difference(filemodel_dict.keys())
        if not to_score:
            #print next(pdb_mdl_dict.iteritems(), None)
            #print next(filemodel_dict.iteritems(), None)
            logging.info("ITScore job completed")
            return fn
        logging.info("ITScoring files...")
        for mdl in to_score:
            modelpath = pdb_mdl_dict[mdl]
            # TODO would it be faster to pass multiple files?
            out, error = self._get_itscore(modelpath)
            # Files with no atoms produce neither out nor error
            if error or not out:
                errors.append((mdl, error))
                score = 0
            else:
                score = out.split()[-1]
            filemodel_dict[mdl] = score

        with open(fn, "w") as oh:
            for mdl in sorted(filemodel_dict.keys()):
                oh.write("%s %s\n" % (mdl, filemodel_dict[mdl]))

        if errors:
            logging.error("%s errors", len(errors))
            for err in errors:
                logging.error(err)
        else:
            logging.info("ITScore job complete.")
            return fn

    @classmethod
    def load(cls, filename):
        names = ("model", "itscore")
        types = (str, float)
        kwargs = dict(names=names,
                      dtype=zip(names, types),
                      delim_whitespace=True)
        mydf = pd.read_csv(filename, **kwargs)
        return mydf

    def _get_itscore(self, model):
        score = subprocess.Popen([self.itscore_bin, model],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 universal_newlines=True)
        out, err = score.communicate()
        return out, err


class OpusPSPScoreError(RuntimeError):
    """Exception for class OpusPSPScore"""
    def __init__(self, message):
        super(OpusPSPScoreError, self).__init__(message)


class OpusPSPScore(ScoreInterface):

    default_output_file = "opuspsp_score.txt"

    def __init__(self, opuspsp_path, input=None, directory=None, pattern=None, wd=None, output_file=None):
        """CONSTRUCTOR"""

        self.opus_psp_dir = opuspsp_path
        opus_psp_bin = os.path.join(self.opus_psp_dir, "opus_psp")
        if shared.missing(opus_psp_bin):
            raise OpusPSPScoreError("{0} missing".format(opus_psp_bin))
        self.opus_psp_bin = opus_psp_bin
        # NB empty string at end forces required trailing slash
        opus_psp_energy = os.path.join(self.opus_psp_dir, "energy_dir", "")
        if not os.path.isdir(opus_psp_energy):
            raise OpusPSPScoreError("{0} missing".format(opus_psp_energy))
        self.opus_psp_energy = opus_psp_energy

        if wd is None:
            wd = self.default_wd

        if pattern is None:
            pattern = self.default_pattern

        filenames = self.get_pdbs(input=input, directory=directory,
                                  pattern=pattern, wd=wd)

        if output_file is None:
            output_file = self.default_output_file
        output_file = os.path.join(wd, output_file)
        logging.info(wd)
        logging.info("%s files for OpusPSP", len(filenames))
        self.run(pdbfiles=filenames, fn=output_file)

    def run(self, pdbfiles, fn=None):
        if fn is None:
            fn = self.default_output_file
        if os.path.isfile(fn) and os.path.getsize(fn):
            # List models already in file
            try:
                opus_df = ScoreShared.load_opuspsp(fn)
            except StopIteration:
                logging.warning("opus psp file had no rows")
                os.remove(fn)
            else:
                filemodels = {os.path.basename(m)[0] for m in opus_df['model'].tolist()}
                # List models in PDB argument
                pdb_mdl_dict = {os.path.basename(m)[0]: m for m in pdbfiles}
                # Find PDBs missing from file
                to_score = set(pdb_mdl_dict.keys()).difference(filemodels)
                if not to_score:
                    logging.info("OpusPSP job completed")
                    return
        logging.info("Scoring files...")

        file_path, rel_files = self.get_relative_pdbs(pdbfiles)

        # Try to write config.psp to file_path
        config_dest = os.path.join(file_path, "config.psp")
        try:
            with open(config_dest, "w") as oh:
                oh.write(self.opus_psp_energy + "\n")
        except IOError as exc:
            if exc.errno == errno.EACCES:
                # If it fails, run in scratch
                logging.error("No write access to decoy directory")
                raise NotImplementedError
            else: raise
        else:
            cmd = [self.opus_psp_bin]
            inp_str = "\n".join(rel_files)
            with open(fn, "w") as oh, shared.CHDIR(file_path):
                proc = subprocess.Popen(cmd,
                                        stdout=oh,
                                        stdin=subprocess.PIPE)
                out, err = proc.communicate(inp_str)
                if proc.returncode:
                    logging.error("OpusPSP returned %s", proc.returncode)
                else:
                    logging.info("OpusPSP complete")
                #os.remove(config_dest)

    @classmethod
    def load(cls, filename):
        names = ("model.pdb", "opuspsp", "opuspsp_od", "opuspsp_ljr")
        types = (str, float, float, float)
        colspecs = [(0, 25), (25, 37), (37, 49), (49, 61)]
        skiprows = 9
        kwargs = dict(names=names,
                      #dtype=zip(names, types),
                      #delim_whitespace=True,
                      colspecs=colspecs,
                      skiprows=skiprows)
        #with open(filename) as ih:
            #for x, line in enumerate(ih):
                #if x >= skiprows: break
                ##print line,
                #logging.debug("skipped row: '%s'", line)
        mydf = pd.read_fwf(filename, **kwargs)
        for n, t in zip(names, types):
            if t == float:
                mydf[n] = mydf[n].astype(t)
        mydf['model'] = mydf.pop("model.pdb").str.replace(".pdb", "")
        return mydf


class MolMechScoreError(RuntimeError):
    """Exception for class MolMechScore"""
    def __init__(self, message):
        super(MolMechScoreError, self).__init__(message)


class MolMechScore(ScoreInterface):

    charmm_f = ("charmm_atom_types.in", "charmm_params.in")

    default_output_file = "molmech_score.txt"

    def __init__(self, lzerd_path, directory=None, input=None, pattern=None, wd=None, output_file=None):
        """CONSTRUCTOR"""

        self.lzerd_dir = lzerd_path
        score_bin = os.path.join(self.lzerd_dir, "score")
        if shared.missing(score_bin):
            raise MolMechScoreError("{0} missing".format(score_bin))
        self.score_bin = score_bin

        if wd is None:
            wd = self.default_wd

        if pattern is None:
            pattern = self.default_pattern

        filenames = self.get_pdbs(input=input, directory=directory,
                                         pattern=pattern, wd=wd)

        if output_file is None:
            output_file = self.default_output_file
        output_file = os.path.join(wd, output_file)

        with shared.CHDIR(wd):
            # Copy charmm.in files
            for f in self.charmm_f:
                dest = os.path.join(wd, f)
                try:
                    os.symlink(os.path.join(self.lzerd_dir, f), dest)
                except OSError as exc:
                    if exc.errno == errno.EEXIST and os.path.isfile(dest):
                        pass
                    else: raise
                except IOError as exc:
                    if exc.errno == errno.EACCES:
                        # If it fails, run in scratch
                        logging.error("No write access to decoy directory")
                        raise NotImplementedError
                    else: raise
            if os.path.isfile(output_file) and os.path.getsize(output_file):
                logging.info("Mol mech score already done")
            else:
                outlines = list()
                for fn in filenames:
                    score = self.run(fn)
                    model = os.path.splitext(os.path.basename(fn))[0]
                    outlines.append(dict(model=model, molmech=score))

                pd.DataFrame(outlines).to_csv(output_file, index=False)

    def run(self, filename):

        score_bin = os.path.expanduser(self.score_bin)

        # Run score
        cmd = [score_bin, "--decoy", filename, "--weights", "all"]
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        retcode = proc.returncode
        if retcode:
            logging.warning("Score %s: %s", filename, retcode)
            return
        for line in out.splitlines():
            if line.startswith("Score"): continue
            parts = line.split(",")
            # Extract score
            score = parts[0]
            break
        return score

    @classmethod
    def load(cls, filename):
        names = ("model", "molmech")
        types = (str, float)
        kwargs = dict(dtype=zip(names, types))
        mydf = pd.read_csv(filename, **kwargs)
        return mydf


class SoapPPScoreError(RuntimeError):
    """Exception for class SoapPPScore"""
    def __init__(self, message):
        super(SoapPPScoreError, self).__init__(message)


class SoapPPScore(ScoreInterface):

    #soap_pp_bin = os.path.normpath(os.path.join(script_dir, os.pardir, "bin", "soap_pp", "soap_pp.py"))

    default_output_file = "soap_pp_score.txt"

    def __init__(self, modeller_cmd, soappp_cmd, input=None, directory=None, pattern=None, wd=None, output_file=None):
        """CONSTRUCTOR"""
        self.modeller_cmd = modeller_cmd
        self.soap_pp_bin = soappp_cmd
        if shared.missing(self.soap_pp_bin):
            raise SoapPPScoreError("{0} missing".format(self.soap_pp_bin))

        if input is not None or (pattern is not None and pattern != "*.pdb"):
            logging.warning("SOAP-PP scores all PDB files in a given directory")
       
        if wd is None:
            wd = self.default_wd

        if pattern is None:
            pattern = self.default_pattern

        filenames = self.get_pdbs(input=input, directory=directory,
                                  pattern=pattern, wd=wd)
        file_path, rel_files = self.get_relative_pdbs(filenames)

        if output_file is None:
            output_file = self.default_output_file
        output_file = os.path.join(wd, output_file)
        logging.info(wd)
        logging.info("%s files for SoapPP", len(filenames))
        #self.run(pdbfiles=filenames, fn=output_file)
        self.run(file_path=file_path, fn=output_file)

    #def run(self, pdbfiles, fn=None):
    def run(self, file_path, fn=None):
        if fn is None:
            fn = self.default_output_file
        if os.path.isfile(fn) and os.path.getsize(fn):
            ## List models already in file
            #soap_df = self.load(fn)
            #filemodels = {os.path.basename(m)[0] for m in soap_df['model'].tolist()}
            ## List models in PDB argument
            #pdb_mdl_dict = {os.path.basename(m)[0]: m for m in pdbfiles}
            ## Find PDBs missing from file
            #to_score = set(pdb_mdl_dict.keys()).difference(filemodels)
            #if not to_score:
                #logging.info("SoapPP job completed")
                #return
            logging.info("SoapPP job completed")
            return
        logging.info("Scoring files...")

        #file_path, rel_files = self.get_relative_pdbs(pdbfiles)

        # Handle space in modeller_cmd (e.g. "modpy.sh python")
        modeller_cmd = self.modeller_cmd
        if " " in modeller_cmd:
            modeller_cmd = modeller_cmd.split()
        else:
            modeller_cmd = [modeller_cmd]
        #cmd = modeller_cmd + [self.soap_pp_bin, "-o", fn]
        cmd = modeller_cmd + [self.soap_pp_bin, file_path]

        #if len(rel_files) < 1000:
            #cmd.extend(rel_files)
        #else:
            #exts = {os.path.splitext(fn)[1] for fn in rel_files}
            #if len(exts) != 1:
                #raise SoapPPScoreError("Many files with more than one file extension")
            #ext = exts.pop()
            #common_name = os.path.commonprefix(rel_files)
            #glob_pattern = "{0}*{1}".format(common_name, ext)
            #cmd.append(glob_pattern)

        # FIXME module? shell=True?
        #cmd = "module load modeller/9.18; {0}".format(subprocess.list2cmdline(cmd))

        # suppress output unless verbose
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            stdout = None
        else:
            stdout = open(os.devnull, "w")
        with shared.CHDIR(file_path):
            ret = subprocess.call(cmd,
                                  stdout=stdout,
                                  stderr=subprocess.STDOUT)
            #ret = subprocess.call(cmd,
                                  #shell=True, executable="/bin/bash",
                                  #stdout=stdout,
                                  #stderr=subprocess.STDOUT,
                                  #close_fds=True,
                                 #)
        if ret:
            logging.error("SoapPP returned %s", ret)
        else:
            outfile = os.path.join(file_path, "soap_pp.score")
            if os.path.isfile(outfile):
                shutil.copyfile(outfile, fn)
                logging.info("SoapPP complete")

    @classmethod
    def load(cls, filename):
        names = ("model:", "soappp")
        types = (str, float)
        kwargs = dict(names=names,
                      dtype=zip(names, types),
                      delim_whitespace=True)
        mydf = pd.read_csv(filename, **kwargs)
        # Handle either ".pdb" or ".pdb:"
        mydf['model'] = mydf.pop("model:").str.replace(":", "")
        mydf['model'] = mydf["model"].str.replace(".pdb", "")
        return mydf


class ShapeScore(ScoreInterface):

    def __init__(self, **kwargs):
        "Do nothing because shape score is in X-Y.out"
        pass

    @staticmethod
    def load(filename):
        """Load shape score and assign model index by shape score rank"""
        names = ("mat0", "mat1", "mat2", "mat3", "mat4", "mat5", "mat6",
                 "mat7", "mat8", "mat9", "mat10", "mat11", "shapescore")
        types = [float] * 13
        #names = ("shape",)
        #types = (float,)
        kwargs = dict(names=names,
                      dtype=zip(names, types),
                      usecols=[12],
                      header=None,
                      delim_whitespace=True)
        mydf = pd.read_table(filename, **kwargs)
        mydf["model"] = "model" + mydf["shapescore"].rank(method="first", ascending=False).astype(int).astype(str)
        return mydf

    @classmethod
    def load_old(cls, filename):
        """Load shape score and assign model index by shape score rank"""
        names = ("shapescore", "shapescorez_old")
        types = (float, float)
        kwargs = dict(names=names,
                      dtype=zip(names, types))
        mydf = pd.read_table(filename, **kwargs)
        # NB pandas rank is a float because average ranks may exist
        mydf["model"] = "model" + mydf["shapescore"].rank(method="first", ascending=False).astype(int).astype(str)
        return mydf


# Key: score
# Value: score object
scorers = dict(
    dfire=GoapScore,
    goap=GoapScore,
    itscore=ITScore,
    opuspsp=OpusPSPScore,
    soappp=SoapPPScore,
    shapescore=ShapeScore,
    molmech=MolMechScore,
    sum=None
)
