#!/usr/bin/env python

# Copyright (C) 2017 Lenna X. Peterson, Daisuke Kihara, and Purdue University
# This file is part of Path-LZerD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from contextlib import contextmanager
import errno
import glob
import inspect
import logging
import os
import shutil
import subprocess
import tempfile

import shared

script_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))


class TemporaryFile(object):
    "Context manager for atomic write"

    def __init__(self, final_path):
        tmpfile_dir = os.path.dirname(final_path)

        self.tmpfile = tempfile.NamedTemporaryFile(dir=tmpfile_dir)
        self.final_path = final_path

    def __getattr__(self, attr):
        return getattr(self.tmpfile, attr)

    def __enter__(self):
        self.tmpfile.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.tmpfile.delete = False
            result = self.tmpfile.__exit__(exc_type, exc_val, exc_tb)
            os.rename(self.tmpfile.name, self.final_path)
            # NamedTemporaryFiles are always 0600
            os.chmod(self.final_path, 0o644)
        else:
            result = self.tmpfile.__exit__(exc_type, exc_val, exc_tb)
        return result


class LzerdPdbgenError(RuntimeError):
    """Exception for class LzerdPdbgen"""
    def __init__(self, message):
        super(LzerdPdbgenError, self).__init__(message)

@contextmanager
def LzerdPdbgen(lzerd, receptor, ligand, number=None, outdir=None, delete=False):
    """Call PDBGEN and create decoys"""

    config = shared.load_config()
    lzerd_path = config['lzerd_path']
    pdbgen_bin = os.path.join(lzerd_path, 'PDBGEN')
    lzerd_data_files = ("charmm_atom_types.in", "charmm_params.in", "uniCHARMM")
    lzerd_data_src = [os.path.join(lzerd_path, fn) for fn in lzerd_data_files]

    errors = list()
    for fn in [lzerd, receptor, ligand, pdbgen_bin] + lzerd_data_src:
        if not os.path.isfile(fn) or not os.path.getsize(fn):
            errors.append(fn)
    if errors:
        logging.error("File(s) missing:")
        for fn in errors:
            logging.error(fn)
        raise LzerdPdbgenError("Required files missing")

    if number is None:
        number = 9999

    if outdir is None:
        outdir = "decoys"
    try:
        os.mkdir(outdir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(outdir):
            pass
        else: raise

    with open(lzerd) as ih:
        number = min(sum(1 for line in ih), number)

    n_decoy_models = len(glob.glob(os.path.join(outdir, "model*.pdb")))
    if n_decoy_models >= number:
        logging.info("Model files complete")
    else:
        # Check whether data files are present, if not copy
        for fn in lzerd_data_src:
            dest = os.getcwd()
            destfile = os.path.join(dest, os.path.basename(fn))
            if not os.path.exists(destfile):
                os.symlink(fn, destfile)
        pdbgen_cmd = [pdbgen_bin,
                      receptor, ligand, lzerd, str(number)]
        devnull = open(os.devnull, "w")
        subprocess.check_call(pdbgen_cmd, stdout=devnull, stderr=devnull)

        with open(receptor) as ih:
            receptor_list = [line for line in ih if line[:4] == "ATOM"]
        pdb_line_fmt = "{:<80}\n"
        ter_line = pdb_line_fmt.format("TER")
        end_line = pdb_line_fmt.format("END")
        receptor_list.append(ter_line)

        for x in xrange(1, number + 1):
            ligand_file = "ligand{}.pdb".format(x)
            model_file = "model{}.pdb".format(x)
            model_file = os.path.join(outdir, model_file)
            # Copy list
            outlist = list(receptor_list)
            with open(ligand_file) as ih:
                outlist.extend(ih)
            outlist.append(ter_line)
            outlist.append(end_line)
            with TemporaryFile(model_file) as oh:
                oh.writelines(outlist)
            os.remove(ligand_file)
    yield
    if delete:
        shutil.rmtree(outdir)

def commandline(module_args=None):
    desc = """Generate LZerD decoys"""
    a = argparse.ArgumentParser(description=desc)
    a.add_argument("lzerd",
                   help="LZerD output (e.g. A-B.out)")
    a.add_argument("-r", "--receptor",
                   help="Receptor PDB file (e.g. A-pdbid.pdb)")
    a.add_argument("-l", "--ligand",
                   help="Ligand PDB file (e.g. B-pdbid.pdb)")
    a.add_argument("-n", "--number", type=int,
                   help="Number of predictions to keep")
    a.add_argument("-o", "--outdir",
                   help="Subdirectory for decoys")
    a.add_argument("-v", "--verbose", action="store_true",
                   help="Print debugging statements")

    args = a.parse_args(module_args)
    kwargs = vars(args)

    level = logging.INFO
    if kwargs.pop("verbose"):
        level = logging.DEBUG
    logging.basicConfig(level=level)

    with LzerdPdbgen(delete=False, **kwargs):
        pass


if __name__ == "__main__":
    commandline()
