#!/usr/bin/env python

# Copyright (C) 2017 Lenna X. Peterson, Daisuke Kihara, and Purdue University
# This file is part of Path-LZerD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import argparse
import copy
import inspect
import itertools
import logging
import os
import StringIO
import subprocess
import traceback

import numpy as np
import pandas as pd
pd.set_option("max_colwidth", 100)

from score_pair import ScorePair
from SCORES import get_score_dict, possible_scores, scorers
import shared


script_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))


class DockingOrderError(RuntimeError):
    """Exception for class DockingOrder"""
    def __init__(self, message):
        super(DockingOrderError, self).__init__(message)


class DockingOrder(object):

    pair_sep = ";"
    item_sep = ","
    chain_sep = "-"

    out_fmt = "{0}-{1}.out"

    default_scores = ["all"]

    @property
    def pathway_scores(self):
        return {scorename + scoretype: [scorename + scoretype, scorename + "z"]
                for scoretype in ["rank"]
                for scorename in self.score_list
               }

    def __init__(self, scores=None):
        """Read configuration"""
        allow_missing = False
        if scores is None:
            logging.info("No scores specified, trying all scores")
            allow_missing = True
            scores = self.default_scores

        config = shared.load_config()
        self.multilzerd_distribution = config['multilzerd_path']

        # For each score, check whether necessary files are found
        if "all" in scores:
            scores = possible_scores.keys()

        new_scores = list()
        for scorename in scores:
            try:
                config_key = possible_scores[scorename]
            except KeyError:
                raise DockingOrderError("Unknown score %s. Scores must be included in SCORES.possible_scores." % scorename)
            else:
                # Keep cases where config_key is None (i.e. no requirements)
                if config_key is None:
                    new_scores.append(scorename)
                    continue
                if isinstance(config_key, (str, unicode)):
                    config_keys = [config_key]
                else:
                    config_keys = config_key
                messages = [None] * len(config_keys)
                for x, config_key in enumerate(config_keys):
                    config_val = config.get(config_key)
                    # Check first part of multi-part cmds
                    if " " in config_val:
                        config_val = config_val.split()[0]
                    # Keep cases where config_key is not None, config_val is not None,
                    # and config_val exists (i.e. requirements met)
                    if config_key is not None and config_val is not None and os.path.exists(config_val):
                        break
                        #new_scores.append(scorename)
                        #continue
                    # Error or skip if config_val is None or config_val does not exist
                    message = None
                    if config_key is not None:
                        if config_val is None:
                            message = "PATHS.ini key '{0}' for score {1} missing".format(config_key, scorename)
                        # Check whether the path doesn't exist or can't be found by `which` (exit code 1)
                        elif not os.path.exists(config_val) and subprocess.call(["which", config_val]):
                            message = "PATHS.ini path '{0}' for score {1} not found".format(config_val, scorename)
                    messages[x] = message
                for m in messages:
                    if m is not None:
                        if allow_missing:
                            logging.info("Skipping: %s", m)
                        else:
                            raise DockingOrderError(m)
                else:
                    new_scores.append(scorename)
        non_sum = [s for s in new_scores if s != "sum"]
        if len(non_sum) == 1:
            logging.error("Cannot use sum with only 1 score, skipping")
            new_scores = non_sum
        logging.info("Using scores: %s", " ".join(new_scores))
        self.score_list = new_scores

    def run_pdb(self, pdbid, top_dir=None, ga_file=None, same_chains=None, output=None, **kwargs):
        """
        Run a single complex
        Tweaked for threading
        """
        if top_dir is None:
            top_dir = os.getcwd()
        if ga_file is None:
            ga_file = os.path.join(top_dir, "{0}.ga.out".format(pdbid))

        if shared.missing(ga_file):
            raise DockingOrderError("Could not find GA file %s" % ga_file)

        pdbid = pdbid.lower()
        #PDBID = pdbid.upper()
        if output is None:
            output = "{0}_path_lzerd.csv".format(pdbid)
        output = os.path.join(top_dir, output)
        outfiles = [output]
        if all(os.path.isfile(f) and os.path.getsize(f) for f in outfiles):
            logging.info("%s done", pdbid)
            return

        try:
            if same_chains is not None:
                same_chains = [list(val) for val in same_chains.split(";")]
            final_kwargs = dict(top_dir=top_dir,
                                ga_file=ga_file,
                                same_chains=same_chains)
            final_predictions = self.predict_case(**final_kwargs)
        except Exception:
            logging.error("Exception for %s", pdbid)
            traceback.print_exc()
        else:
            pred_df_dict = {"Final generation": final_predictions['pathways']}

            all_method_df_list = list()
            for method, pred_df in pred_df_dict.items():
                highest_counts = pred_df.idxmax(axis=1)
                count_frame = highest_counts.to_frame("pathway")
                #count_frame['method'] = method
                count_frame['pdbid'] = pdbid
                all_method_df_list.append(count_frame)
            pdb_df = pd.concat(all_method_df_list, ignore_index=False)
            pdb_df.index.names = ["score"]
            pdb_df.to_csv(output)
            return

    def predict_case(self, top_dir, ga_file, same_chains=None):
        "Predict assembly order using votes from Multi-LZerD GA file"
        # Overall workflow:
        # Load multi decoys from *.ga.out
        # Load pairwise scores
        # Compute each decoy's paths based on scores/score ranks
        # (generate pairwise decoys if missing)
        # re-score pairwise decoys
        # Convert paths to pathways
        # Count pathways per score

        # NB if ga_file is an absolute path, top_dir will be ignored
        ga_header, ga_df = self.read_ga_file(os.path.join(top_dir, ga_file))
        chain_list = ga_header['Chains'].split(self.item_sep)
        all_chains_df = self.read_score_files(top_dir=top_dir, chain_list=chain_list, pdbid=ga_header['PDBID'])

        ga_df_convert = ga_df.apply(self.convert_decoylist, all_chains_df=all_chains_df, chain_list=chain_list, axis=1)

        pathways = list()
        for pathway_score_name in self.pathway_scores.keys():
            pair_key = pathway_score_name + "_pairs"

            pathway_key = "pathway" + "_" + pathway_score_name
            ga_df_convert[pathway_key] = ga_df_convert.pop(pair_key).apply(
                lambda x: self.repr_pathway(self.pairs_to_pathway(x)))
            if same_chains:
                ga_df_convert[pathway_key] = ga_df_convert[pathway_key].apply(self.reletter_pathway, same_chains=same_chains)
            pathway_counts = ga_df_convert.groupby(pathway_key).apply(len)
            pathway_sorted = pathway_counts.sort_values(ascending=False)
            pathway_df = pathway_sorted.to_frame(pathway_score_name)
            pathways.append(pathway_df.T)
        ret = dict(pathways=pd.concat(pathways))
        return ret

    @staticmethod
    def reletter_pathway(pathway, same_chains):
        "Use the same letter to represent multiple copies of the same protein"
        for chain_list in same_chains:
            first_chain = chain_list[0]
            for replace_chain in chain_list[1:]:
                pathway = pathway.replace(replace_chain, first_chain)
        pathway_tmp = list()
        steps = pathway.split(">")
        for st in steps:
            subcomplexes = st.split("+")
            if len(subcomplexes) == 1:
                new_step = "".join(sorted(st))
            else:
                subcomplex_tmp = list()
                for sc in subcomplexes:
                    subcomplex_tmp.append("".join(sorted(sc)))
                new_step = "+".join(sorted(subcomplex_tmp))
            pathway_tmp.append(new_step)
        return ">".join(pathway_tmp)

    ga_header_len = 8
    ga_names = ["decoylist", "score1", "score2", "score3"]
    @classmethod
    def read_ga_file(cls, ga_file):
        "Read Multi-LZerD genetic algorithm file"
        # Read data from top of file
        ga_header_len = 8
        ga_header = dict()
        with open(ga_file) as ih:
            for x in range(ga_header_len):
                line = next(ih)
                parts = line.strip().split(None, 1)
                if parts:
                    ga_header[parts[0]] = parts[1]

        # Read rest of file into data frame
        with open(ga_file) as ih:
            ga_df = pd.read_table(ih, delim_whitespace=True, 
                                  skiprows=ga_header_len,
                                  names=cls.ga_names)

        return ga_header, ga_df

    @classmethod
    def get_lzerd_files(cls, top_dir, chain_list, pdbid):
        "Find pairwise files for LZerD"
        pdb_fmt = os.path.join(top_dir, "{0}-%s.pdb" % pdbid)
        for chain_pair in itertools.combinations(chain_list, 2):
            pdb_kwargs = {kind: pdb_fmt.format(chain_pair[x])
                          for x, kind in enumerate(["receptor", "ligand"])}
            lzerd_output = cls.out_fmt.format(*chain_pair)
            pdb_kwargs['lzerd_output'] = os.path.join(top_dir, lzerd_output)
            pdb_kwargs['chain_pair'] = chain_pair
            yield pdb_kwargs

    def read_score_files(self, top_dir, chain_list, pdbid):

        all_chains_df_list = list()

        for row in self.get_lzerd_files(top_dir=top_dir, chain_list=chain_list, pdbid=pdbid):
            out_fn = row['lzerd_output']
            chain_pair = row.pop('chain_pair')
            chain_df_dict = dict()

            with open(out_fn) as ih:
                n_decoys = sum(1 for __ in ih)
            # Run scores
            # NB dfire, goap, opus, soap, itscore give same scores with/without H
            score_kwargs = dict(number=54000,
                                dest_dir=top_dir,
                                scores=self.score_list,
                                **row)
            ScorePair(**score_kwargs)
            # Always get shapescore
            for scorename in set(self.score_list + ["shapescore"]):
                # Skip sum
                if scorename == "sum":
                    continue
                score_type = get_score_dict(scorename)
                filename = score_type['fmt'].format(*chain_pair)
                fn = os.path.join(top_dir, filename)
                if not os.path.isfile(fn):
                    logging.error("%s missing", fn)
                    continue
                score_df = self.load_score(filename=fn, scorename=scorename)
                if len(score_df) < n_decoys:
                    logging.error("Too few rows from %s: %s", fn, len(score_df))
                    os.remove(fn)
                    continue
                multiplier = 1.0
                if not score_type['ascending']:
                    multiplier = -1.0
                # Compute Z-score of score
                score_df[scorename + "z"] = self.zscore(multiplier * score_df[scorename])
                # Compute rank of score and cast to int
                score_df[scorename + "rank"] = score_df[scorename].rank(method="first", ascending=score_type['ascending']).astype(int)
                # Parse modelx row
                score_df['modelid'] = score_df.apply(lambda r: int(r["model"].replace("model", "")), axis=1)
                # Fix duplicate model_x model_y columns
                del score_df['model']
                chain_df_dict[scorename] = score_df
            if not chain_df_dict:
                logging.error("No scores found for %s", chain_pair)
                continue
            ## Combine scores
            all_score_df = chain_df_dict.pop("shapescore")
            # Merge other dfs
            for other_df in chain_df_dict.values():
                all_score_df = pd.merge(all_score_df, other_df,
                                    #on="modelid",
                                    how="left", sort=False)
            # Compute sum
            if "sum" in self.score_list:
                normal_scores = [s for s in self.score_list if s != "sum"]
                assert len(normal_scores) > 1
                for suffix in "rank", "z":
                    cols = [s + suffix for s in normal_scores]
                    all_score_df["sum" + suffix] = all_score_df[cols].sum(axis=1)
            chain_id = self.chain_sep.join(chain_pair)
            all_score_df['chain_pair'] = chain_id
            all_chains_df_list.append(all_score_df)
        if not all_chains_df_list:
            logging.error("No pairwise scores found in %s", top_dir)
            raise DockingOrderError("No pairwise scores for %s" % top_dir)
        all_chains_df = pd.concat(all_chains_df_list, ignore_index=True)

        return all_chains_df

    @classmethod
    def load_score(cls, filename, scorename):
        obj = scorers.get(scorename)
        if obj is None:
            logging.warning("No score loading method for %s", scorename)
            return pd.DataFrame()
        return obj.load(cls.filter_file(filename))

    ignore_lines = ["nohup", "Writing data to"]
    @classmethod
    def filter_file(cls, fn):
        """Remove bad lines from files"""
        with open(fn) as ih:
            ret = StringIO.StringIO("".join(line for line in ih
                                            if not any(line.startswith(ignore)
                                                       for ignore in cls.ignore_lines))
                                   )
        return ret

    @staticmethod
    def zscore(s):
        return (s - np.mean(s)) / np.std(s)

    def convert_decoylist(self, row, all_chains_df, chain_list):
        decoylist = row['decoylist']
        # Split list of decoys by ; into pairs
        pair_list = decoylist.split(self.pair_sep)
        # Split each pair by , into list of first chain index, second chain index, shape score index
        pair_lol = [pair.split(self.item_sep) for pair in pair_list]
            
        # Convert chain index to chain letter and join with -
        converted_lol = [[chain_list[int(sublist[0])] + self.chain_sep + chain_list[int(sublist[1])], sublist[2]]
                         for sublist in pair_lol]

        return self.convert_ranks(row=row, converted_lol=converted_lol, all_chains_df=all_chains_df)
        
    def convert_ranks(self, row, converted_lol, all_chains_df):
        # Make data frame
        decoy_df = pd.DataFrame(converted_lol)
        decoy_df.columns = ["chain_pair", "shapescoreindex"]
        # Convert 0-based index to 1-based rank
        decoy_df['shapescorerank'] = pd.to_numeric(decoy_df.pop("shapescoreindex")) + 1
        # Join with master list
        decoy_df = pd.merge(decoy_df, all_chains_df, on=['chain_pair', "shapescorerank"])
        n_rows = len(decoy_df)
        for scorename, sort_keys in self.pathway_scores.items():
            scorename_key = scorename + "_pairs"
            subset_df = decoy_df.dropna(subset=sort_keys)
            if subset_df.empty or len(subset_df) < n_rows:
                row[scorename_key] = np.nan
            else:
                row[scorename_key] = tuple(decoy_df.sort_values(sort_keys)['chain_pair'].tolist())
        
        # Re-join chain pair and shape score index with ,
        converted_list = [self.item_sep.join(lol) for lol in converted_lol]
            
        # Re-join pairs into decoy list with ;
        row['decoyletter'] = self.pair_sep.join(converted_list)
        return row

    @staticmethod
    def is_simple(pathway):
        return all(isinstance(v, basestring) for v in pathway)

    @classmethod
    def pairs_to_pathway(cls, pairs):
        # Initialize pathway with starting pair
        pathway = dict()
        pathway[1] = sorted(pairs[0].split(cls.chain_sep))
        
        for x, pr in enumerate(pairs[1:]):
            n = x + 1
            current_pathway = pathway[n]
            current_length = len(pathway[n])
            is_simple = cls.is_simple(current_pathway)
            pathway_str = str(current_pathway)
            chains = pr.split(cls.chain_sep)
            chain1, chain2 = chains
            # case 1: two existing complexes can be connected by the new pair
            if chain1 in pathway_str and chain2 in pathway_str:
                assert not is_simple
                if current_length == 2:
                    # If there are exactly 2 subcomplexes, the new pathway is everything combined
                    new_pathway = sorted(itertools.chain(*current_pathway))
                elif current_length > 2:
                    # Otherwise, only combine the matched subcomplexes
                    unused_subcomplexes = list()
                    used_subcomplexes = list()
                    for subcomplex in current_pathway:
                        current_subcomplex = list(subcomplex)
                        if chain1 in subcomplex or chain2 in subcomplex:
                            used_subcomplexes.append(current_subcomplex)
                        else:
                            unused_subcomplexes.append(current_subcomplex)
                    assert len(used_subcomplexes) == 2, len(used_subcomplexes)
                    new_subcomplex = sorted(itertools.chain(*used_subcomplexes))
                    new_pathway = [new_subcomplex]
                    new_pathway.extend(unused_subcomplexes)
                else:
                    assert False
            # case 2: the new pair can be connected to one pre-existing subcomplex
            elif chain1 in pathway_str or chain2 in pathway_str:
                if chain1 in pathway_str:
                    matched_chain = chain1
                    other_chain = chain2
                elif chain2 in pathway_str:
                    matched_chain = chain2
                    other_chain = chain1
                    
                if is_simple:
                    current_complex = list(current_pathway)
                    current_complex.append(other_chain)
                    new_pathway = sorted(current_complex)
                else:
                    new_pathway = list()
                    for m, subcomplex in enumerate(current_pathway):
                        current_subcomplex = list(subcomplex)
                        if matched_chain in subcomplex:
                            current_subcomplex.append(other_chain)
                            current_subcomplex.sort()
                        new_pathway.append(current_subcomplex)

            # case 3: the new pair cannot be connected to any existing complex
            elif chain1 not in pathway_str and chain2 not in pathway_str:
                new_subcomplex = list(chains)
                if is_simple:
                    new_pathway = [list(current_pathway), new_subcomplex]
                else:
                    new_pathway = copy.deepcopy(current_pathway)
                    new_pathway.append(new_subcomplex)
            else:
                assert False
            
            pathway[n + 1] = new_pathway
        
        return pathway

    @classmethod
    def repr_pathway(cls, pathway):
        parts = list()
        for k in sorted(pathway.keys()):
            step = pathway[k]
            if cls.is_simple(step):
                step_repr = "".join(step)
            else:
                step_repr = "+".join("".join(subcomplex) for subcomplex in sorted(step))
            parts.append(step_repr)
        return ">".join(parts)

    @classmethod
    def test_pairs_to_pathway(cls):
        pairs_to_pathway_input = ['B-G', 'E-F', 'B-E', 'C-D', 'A-D', 'C-G']
        pairs_to_pathway_output = {1: ['B', 'G'], 2: [['B', 'G'], ['E', 'F']], 3: ['B', 'E', 'F', 'G'], 4: [['B', 'E', 'F', 'G'], ['C', 'D']], 5: [['B', 'E', 'F', 'G'], ['A', 'C', 'D']], 6: ['A', 'B', 'C', 'D', 'E', 'F', 'G']}

        test_output = cls.pairs_to_pathway(pairs_to_pathway_input)
        assert test_output == pairs_to_pathway_output, test_output

    @classmethod
    def test(cls):
        cls.test_pairs_to_pathway()

    @classmethod
    def commandline(cls, module_args=None):
        desc = """
        Path-LZerD: multimeric protein complex assembly order prediction
        """
        a = argparse.ArgumentParser(description=desc)
        a.add_argument("-p", "--pdbid", required=True,
                       help="ID of PDB (PDB files named A-pdbid.pdb, B-pdbid.pdb etc)")
        a.add_argument("-d", "--top_dir",
                       help="Directory containing LZerD output")
        a.add_argument("-g", "--ga_file",
                       help="Multilzerd output (e.g. pdbid.ga.out)")
        a.add_argument("-c", "--same_chains", metavar="CHAINS[;CHAINS ...]",
                       help="Chains that represent the same protein separated with semicolons (;), e.g. 'AB;CD'")
        a.add_argument("-s", "--scores", nargs="*",
                       choices=["all"] + possible_scores.keys(),
                       help="Scoring function(s) to run")
        a.add_argument("-t", "--test", action="store_true",
                       help="Test pair to pathway generation")
        a.add_argument("-o", "--output",
                       help="Output file")
        a.add_argument("-v", "--verbose", action="store_true",
                       help="Print debugging statements")

        args = a.parse_args(module_args)
        kwargs = vars(args)

        level = logging.INFO
        if kwargs.pop("verbose"):
            level = logging.DEBUG
        logging.basicConfig(level=level)

        test = kwargs.pop("test")
        if test:
            cls.test()
            return
        c = cls(kwargs.pop("scores"))
        c.run_pdb(**kwargs)
        return c


if __name__ == "__main__":
    DockingOrder.commandline()
