#!/usr/bin/env python

# Copyright (C) 2017 Lenna X. Peterson, Daisuke Kihara, and Purdue University
# This file is part of Path-LZerD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import argparse
import errno
import glob
import inspect
import logging
import os
import traceback

from lzerd_pdbgen import LzerdPdbgen
from SCORES import possible_scores, scorers
import shared

script_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))


class ScorePairError(RuntimeError):
    """Exception for class ScorePair"""
    def __init__(self, message):
        super(ScorePairError, self).__init__(message)


class ScorePair(object):

    def __init__(self, lzerd_output, receptor, ligand, scores=None, number=None, dest_dir=None):
        """
        Score a pair of docking decoys.
        """
        if number is None:
            number = 54000

        if dest_dir is None:
            dest_dir = os.path.dirname(lzerd_output)

        if scores is None or scores == ["all"] or scores == "all":
            scores = possible_scores.keys()

        self.config = shared.load_config()

        self.run(lzerd_output=lzerd_output,
                 receptor=receptor,
                 ligand=ligand,
                 scores=scores,
                 number=number,
                 dest_dir=dest_dir)

    def run(self, lzerd_output, receptor, ligand, scores, number, dest_dir):

        lzerd_output = os.path.abspath(lzerd_output)
        receptor = os.path.abspath(receptor)
        ligand = os.path.abspath(ligand)
        dest_dir = os.path.abspath(dest_dir)

        try:
            os.mkdir(dest_dir)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(dest_dir):
                pass
            else: raise

        pair_id = os.path.splitext(os.path.basename(lzerd_output))[0]

        # Check if all score files are done
        score_files = [self.scorefilename(s, id=pair_id, directory=dest_dir)
                       for s in scores]
        if not any(shared.missing(f) for f in score_files):
            return

        decoy_dir = os.path.join(dest_dir, "decoys")
        with shared.CHDIR(dest_dir), LzerdPdbgen(lzerd=lzerd_output, receptor=receptor, ligand=ligand, number=number, outdir=decoy_dir, delete=True):

            files = glob.glob(os.path.join(decoy_dir, "*.pdb"))

            kwargs = dict(wd=dest_dir, input=files)

            for scorename in scores:
                if scorename in ("sum", "shapescore"):
                    continue
                callable = scorers.get(scorename)
                if callable is None:
                    logging.error("No callable for score %s found in SCORES.scorers", scorename)
                    continue
                score_config = possible_scores.get(scorename)
                score_kwargs = dict(kwargs)
                if score_config is not None:
                    if isinstance(score_config, (str, unicode)):
                        score_keys = [score_config]
                    else:
                        score_keys = score_config
                    for k in score_keys:
                        score_kwargs[k] = self.config.get(k)
                filename = self.scorefilename(scorename, id=pair_id)
                file_dest = os.path.join(dest_dir, filename)
                if shared.missing(file_dest):
                    try:
                        callable(output_file=file_dest, **score_kwargs)
                    except Exception:
                        logging.error("%s %s", scorename, callable)
                        traceback.print_exc()
                    if not os.path.isfile(file_dest):
                        logging.error("No output from %s", scorename)
                        continue

    @staticmethod
    def scorefilename(scorename, id=None, directory=None):
        parts = [scorename]
        if id:
            parts.insert(0, id)
        filename = "_".join(parts) + ".txt"
        if directory:
            filename = os.path.join(directory, filename)
        return filename

    @classmethod
    def commandline(cls, module_args=None):
        desc = """Generate pairwise decoys, run score, delete decoys"""
        a = argparse.ArgumentParser(description=desc)
        a.add_argument("lzerd_output",
                       help="LZerD output file (e.g. A-B.out)")
        a.add_argument("-r", "--receptor", required=True,
                       help="Receptor PDB path")
        a.add_argument("-l", "--ligand", required=True,
                       help="Ligand PDB path")
        a.add_argument("-s", "--scores", nargs="*",
                       choices=["all"] + possible_scores.keys(),
                       help="Scoring function(s) to run")
        a.add_argument("-n", "--number", type=int,
                       help="Max number of decoys to generate")
        a.add_argument("-o", "--dest_dir",
                       help="Directory to copy score file to")
        a.add_argument("-v", "--verbose", action="store_true",
                       help="Print debugging statements")

        args = a.parse_args(module_args)
        kwargs = vars(args)

        level = logging.INFO
        if kwargs.pop("verbose"):
            level = logging.DEBUG
        logging.basicConfig(level=level)

        c = cls(**kwargs)
        return c


if __name__ == "__main__":
    ScorePair.commandline()
