# Copyright (C) 2017 Lenna X. Peterson, Daisuke Kihara, and Purdue University
# This file is part of Path-LZerD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import inspect
import os

script_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))
# Root directory of package
ROOTDIR = script_dir


class PathLzerdError(RuntimeError):
    """Exception for Path-LZerD"""
    def __init__(self, message):
        super(PathLzerdError, self).__init__(message)

#### FILE SYSTEM FUNCTIONS ####
class CHDIR(object):
    def __init__(self, dirname):
        self.dirname = dirname

    def __enter__(self):
        self.old_dir = os.getcwd()
        os.chdir(self.dirname)

    def __exit__(self, exception_type, exception_value, traceback):
        os.chdir(self.old_dir)

def missing(*files):
    "Check whether files are missing or empty"
    return any(not os.path.isfile(f) or not os.path.getsize(f)
               for f in files)

def mkdir_p(path):
    """
    Pythonic replacement for shell `mkdir -p`
    Creates multiple directories and ignores error caused by existing
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise


def load_config():
    # Parse paths
    path_file = os.path.join(ROOTDIR, "PATHS.ini")
    if not os.path.isfile(path_file):
        raise PathLzerdError("Could not open %s" % path_file)
    config = dict()
    with open(path_file) as ih:
        for line in ih:
            line = line.strip()
            parts = line.split(":")
            if len(parts) == 2:
                key, val = parts
                key = key.strip()
                val = val.strip()
                config[key] = os.path.expandvars(val)
            else:
                if line[0] != "[" or line[-1] != "]":
                    print "Could not interpret config line %s" % line

    required_keys = ('lzerd_path', 'multilzerd_path')

    errors = 0
    for key in required_keys:
        if key not in config:
            errors += 1
            print "'{0}' did not contain required key '{1}'".format(path_file, key)
        elif not os.path.exists(os.path.expanduser(config[key])):
            errors += 1
            print "Value for {0}: {1} (path not found)".format(key, config[key])
    if not errors:
        return config
    else:
        raise PathLzerdError("{0} error(s) processing {1}".format(errors, path_file))
